import pytest

try:
    from unittest.mock import patch, mock_open
    builtin_open = "builtins.open"
except Exception:
    from mock import patch, mock_open
    builtin_open = "__builtin__.open"

from os import path
from insights.core import filters
from insights.core.exceptions import SkipComponent
from insights.parsers.iris import IrisList, IrisCpf
from insights.specs import Specs
from insights.specs.datasources.intersystems import iris_working_configuration, iris_working_messages_log
from insights.tests import context_wrap


IRIR_LIST = """

Configuration 'IRIS'   (default)
    directory:    /intersystems
    versionid:    2023.1.0.235.1com
    datadir:      /intersystems
    conf file:    iris.cpf  (SuperServer port = 1972, WebServer = 52773)
    status:       running, since Wed Jul 19 02:37:46 2023
    state:        ok
    product:      InterSystems IRIS
""".strip()


IRIS_CPF = """
[ConfigFile]
Product=IRIS
Version=2023.1

[Databases]
IRISSYS=/intersystems/mgr/
IRISLIB=/intersystems/mgr/irislib/
IRISTEMP=/intersystems/mgr/iristemp/
IRISLOCALDATA=/intersystems/mgr/irislocaldata/
IRISAUDIT=/intersystems/mgr/irisaudit/
ENSLIB=/intersystems/mgr/enslib/
USER=/intersystems/mgr/user/

[Namespaces]
%SYS=IRISSYS
USER=USER
""".strip()


RAW_MESSAGES = """
06/26/23-08:02:17:828 (144145) 0 [Generic.Event] Allocated 495MB shared memory
06/26/23-08:02:17:828 (144145) 0 [Generic.Event] 32MB global buffers, 80MB routine buffers, 64MB journal buffers, 4MB buffer descriptors, 300MB heap, 5MB ECP, 9MB miscellaneous
06/26/23-08:02:17:831 (144145) 0 [Crypto.IntelSandyBridgeAESNI] Intel Sandy Bridge AES-NI instructions detected.
06/26/23-08:02:17:831 (144145) 0 [SIMD] SIMD optimization level: DEFAULT X86
06/26/23-08:02:17:903 (144145) 0 [WriteDaemon.UsingWIJFile] Using WIJ file: /intersystems/mgr/IRIS.WIJ
06/26/23-08:02:17:903 (144145) 0 [Generic.Event] No journaling info from prior system
06/26/23-08:02:17:903 (144145) 0 [WriteDaemon.CreatingNewWIJ] Creating a new WIJ file
06/26/23-08:02:18:104 (144145) 0 [WriteDaemon.CreatedNewWIJ] New WIJ file created
06/26/23-08:02:18:110 (144145) 0 [Generic.Event]
""".strip()


FILTERED_MESSAGES = """
06/26/23-08:02:17:828 (144145) 0 [Generic.Event] Allocated 495MB shared memory
06/26/23-08:02:17:828 (144145) 0 [Generic.Event] 32MB global buffers, 80MB routine buffers, 64MB journal buffers, 4MB buffer descriptors, 300MB heap, 5MB ECP, 9MB miscellaneous
06/26/23-08:02:17:903 (144145) 0 [Generic.Event] No journaling info from prior system
06/26/23-08:02:18:110 (144145) 0 [Generic.Event]
""".strip()


def setup_function(func):
    if Specs.iris_messages_log in filters._CACHE:
        del filters._CACHE[Specs.iris_messages_log]
    if Specs.iris_messages_log in filters.FILTERS:
        del filters.FILTERS[Specs.iris_messages_log]

    if func is test_iris_working_messages_log:
        filters.add_filter(Specs.iris_messages_log, ["Generic.Event"])
    if func is test_iris_working_messages_log_no_match_filter:
        filters.add_filter(Specs.iris_messages_log, ["test_no_match_filter"])
    if func is test_iris_working_messages_log_no_filter:
        filters.add_filter(Specs.iris_messages_log, [])


@patch("insights.specs.datasources.intersystems.open", new_callable=mock_open, read_data=IRIS_CPF)
@patch("os.path.isfile", return_value=True)
def test_iris_working_configuration(m_open, m_isfile):
    iris_list_info = IrisList(context_wrap(IRIR_LIST))
    broker = {IrisList: iris_list_info}
    result = iris_working_configuration(broker)
    assert result.content == IRIS_CPF.splitlines()
    assert result.relative_path == path.join(path.dirname(__file__), '/intersystems/iris.cpf')


@patch("insights.specs.datasources.intersystems.open", new_callable=mock_open, read_data=IRIS_CPF)
def test_iris_working_configuration_no_file(m_open):
    iris_list_info = IrisList(context_wrap(IRIR_LIST))
    broker = {IrisList: iris_list_info}
    with pytest.raises(SkipComponent) as e:
        iris_working_configuration(broker)
    assert 'SkipComponent' in str(e)


@patch("insights.specs.datasources.intersystems.open", new_callable=mock_open, read_data=RAW_MESSAGES)
@patch("os.path.isfile", return_value=True)
def test_iris_working_messages_log(m_open, m_isfile):
    iris_cpf_info = IrisCpf(context_wrap(IRIS_CPF))
    broker = {IrisCpf: iris_cpf_info}
    result = iris_working_messages_log(broker)
    assert result.content == FILTERED_MESSAGES.splitlines()
    assert result.relative_path == path.join(path.dirname(__file__), '/intersystems/mgr/messages.log')


@patch("insights.specs.datasources.intersystems.open", new_callable=mock_open, read_data=RAW_MESSAGES)
def test_iris_working_messages_log_no_file(m_open):
    iris_cpf_info = IrisCpf(context_wrap(IRIS_CPF))
    broker = {IrisCpf: iris_cpf_info}
    with pytest.raises(SkipComponent) as e:
        iris_working_messages_log(broker)
    assert 'SkipComponent' in str(e)


@patch("insights.specs.datasources.intersystems.open", new_callable=mock_open, read_data=RAW_MESSAGES)
@patch("os.path.isfile", return_value=True)
def test_iris_working_messages_log_no_match_filter(m_open, m_isfile):
    iris_cpf_info = IrisCpf(context_wrap(IRIS_CPF))
    broker = {IrisCpf: iris_cpf_info}
    result = iris_working_messages_log(broker)
    assert result.content == []
    assert result.relative_path == path.join(path.dirname(__file__), '/intersystems/mgr/messages.log')


@patch("insights.specs.datasources.intersystems.open", new_callable=mock_open, read_data=RAW_MESSAGES)
@patch("os.path.isfile", return_value=True)
def test_iris_working_messages_log_no_filter(m_open, m_isfile):
    iris_cpf_info = IrisCpf(context_wrap(IRIS_CPF))
    broker = {IrisCpf: iris_cpf_info}
    with pytest.raises(SkipComponent) as e:
        iris_working_messages_log(broker)
    assert 'SkipComponent' in str(e)



"""
Custom datasources to get the iris working configuration/log files.
"""
import os

from insights.core.context import HostContext
from insights.core.exceptions import ContentException, SkipComponent
from insights.core.filters import get_filters
from insights.core.plugins import datasource
from insights.core.spec_factory import DatasourceProvider
from insights.parsers.iris import IrisList, IrisCpf
from insights.specs import Specs


@datasource(IrisList, HostContext)
def iris_working_configuration(broker):
    """
    This datasource get information from ``iris.cpf`` file.

    Returns:
        str: the raw content of the ``iris.cpf`` file .

    Raises:
        SkipComponent: When the `iris.cpf` does not exist
        ContentException: When any exception occurs.
    """
    configuration_file_path_directory = broker[IrisList]['directory']
    configuration_file_path_conf_file = broker[IrisList]['conf file'].split()[0].strip()

    configuration_file_path = os.path.join(configuration_file_path_directory, configuration_file_path_conf_file)
    if os.path.isfile(configuration_file_path):
        try:
            with open(configuration_file_path, 'r') as conf:
                return DatasourceProvider(content=conf.read(), relative_path=configuration_file_path)
        except Exception as e:
            raise ContentException(e)
    raise SkipComponent


@datasource(IrisCpf, HostContext)
def iris_working_messages_log(broker):
    """
    This datasource get information from ``messages.log`` file.

    Returns:
        str: the filtered content of the ``messages.log`` file .

    Raises:
        SkipComponent: When the log file/option/filters do not exist
        ContentException: When any exception occurs.
    """
    filters = sorted((get_filters(Specs.iris_messages_log)))
    iris_cpf = broker[IrisCpf]

    if filters and iris_cpf.has_option('Databases', 'IRISSYS'):
        log_path = iris_cpf.get("Databases", "IRISSYS")
        log_path = os.path.join(log_path, 'messages.log')
        if os.path.isfile(log_path):
            try:
                with open(log_path, 'r') as log:
                    filtered_content = []
                    for line in log.readlines():
                        filtered_content.append(line) if any(_f in line for _f in filters) else None
                    return DatasourceProvider(content="".join(filtered_content), relative_path=log_path)
            except Exception as e:
                raise ContentException(e)
    raise SkipComponent



"""
crypto-policies - files in ``/etc/crypto-policies/back-ends/``
==============================================================

This is a collection of parsers that all deal with the generated configuration
files under the ``/etc/crypto-policies/back-ends/`` folder.  Parsers included
in this module are:

CryptoPoliciesConfig - file ``/etc/crypto-policies/config``
-----------------------------------------------------------

CryptoPoliciesStateCurrent - file ``/etc/crypto-policies/state/current``
------------------------------------------------------------------------

CryptoPoliciesOpensshserver - file ``/etc/crypto-policies/back-ends/opensshserver.config``
------------------------------------------------------------------------------------------

CryptoPoliciesBind - file ``/etc/crypto-policies/back-ends/bind.config``
------------------------------------------------------------------------
"""
from insights.core import Parser, SysconfigOptions
from insights.core.exceptions import SkipComponent
from insights.core.plugins import parser
from insights.parsers import get_active_lines
from insights.specs import Specs


@parser(Specs.crypto_policies_config)
class CryptoPoliciesConfig(Parser):
    """
    This parser reads the ``/etc/crypto-policies/config`` file.
    The contents of the file is a single-line value, available
    in the ``value`` property.

    Sample Input::

        LEGACY

    Examples:
        >>> cp_c.value
        'LEGACY'
    """
    def parse_content(self, content):
        if not content:
            raise SkipComponent("/etc/crypto-policies/config is empty")
        self.value = get_active_lines(content)[0]


@parser(Specs.crypto_policies_state_current)
class CryptoPoliciesStateCurrent(Parser):
    """
    This parser reads the ``/etc/crypto-policies/state/current`` file.
    The contents of the file is a single-line value, available
    in the ``value`` property.

    Sample Input::

        LEGACY

    Examples:
        >>> cp_sc.value
        'LEGACY'
    """
    def parse_content(self, content):
        if not content:
            raise SkipComponent("/etc/crypto-policies/state/current is empty")
        self.value = get_active_lines(content)[0]


@parser(Specs.crypto_policies_opensshserver)
class CryptoPoliciesOpensshserver(SysconfigOptions):
    """
    This parser reads the ``/etc/crypto-policies/back-ends/opensshserver.config``
    file.  It uses the ``SysconfigOptions`` parser class to convert the file into
    a dictionary of options. It also provides the ``options`` property as a helper
    to retrieve the ``CRYPTO_POLICY`` variable.

    Sample Input::

        CRYPTO_POLICY='-oCiphers=aes256-gcm@openssh.com,3des-cbc -oMACs=umac-128-etm@openssh.com'

    Examples:
        >>> 'CRYPTO_POLICY' in cp_os
        True
        >>> cp_os.options
        '-oCiphers=aes256-gcm@openssh.com,3des-cbc -oMACs=umac-128-etm@openssh.com'
    """
    @property
    def options(self):
        """ (union[str, None]): The value of the ``CRYPTO_POLICY`` variable if it exists, else None."""
        return self.data.get('CRYPTO_POLICY', None)


@parser(Specs.crypto_policies_bind)
class CryptoPoliciesBind(Parser):
    """
    This parser reads the ``/etc/crypto-policies/back-ends/bind.config`` file.
    The sections ``disable-algorithms`` and ``disable-ds-digests`` are in the
    properties ``disable_algorithms`` and ``disable_ds_digests``.

    Sample Input::

        disable-algorithms "." {
        RSAMD5;
        DSA;
        };
        disable-ds-digests "." {
        GOST;
        };

    Examples:
        >>> 'GOST' in cp_bind.disable_ds_digests
        True
        >>> cp_bind.disable_algorithms
        ['RSAMD5', 'DSA']
    """
    def parse_content(self, content):
        if not content:
            raise SkipComponent("/etc/crypto-policies/back-ends/bind.config is empty")
        self.value = content
        in_da = False
        in_ddd = False
        da = []
        ddd = []
        for line in self.value:
            if line.strip().lower().startswith("disable-algorithms"):
                in_da = True
                continue
            if line.strip().lower().startswith("disable-ds-digests"):
                in_ddd = True
                continue
            if line.strip().startswith("}"):
                in_da = False
                in_ddd = False
                continue
            algo = line.strip().strip(''';'"''')
            if in_da:
                da.append(algo)
            if in_ddd:
                ddd.append(algo)
        self.disable_algorithms = da
        self.disable_ds_digests = ddd
